from django.test import override_settings

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models import Document
from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import TEST_COMPRESSED_DOCUMENT_PATH

from ..source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_ALWAYS

from .mixins.base_mixins import (
    DocumentFileUploadViewTestMixin, DocumentUploadWizardViewTestMixin
)
from .mixins.web_form_source_mixins import WebFormSourceTestMixin


class WebFormDocumentUploadWizardViewTestCase(
    WebFormSourceTestMixin, DocumentUploadWizardViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    @override_settings(DOCUMENTS_LANGUAGE='fra')
    def test_default_document_language_setting(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_upload_interactive_view()
        self.assertContains(
            response=response,
            text='<option value="fra" selected>French</option>',
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_compressed_file(self):
        backend_data = self.test_source.get_backend_data()
        backend_data['uncompress'] = SOURCE_UNCOMPRESS_CHOICE_ALWAYS
        self.test_source.set_backend_data(obj=backend_data)
        self.test_source.save()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_wizard_view(
            document_path=TEST_COMPRESSED_DOCUMENT_PATH
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), test_document_count + 2)
        self.assertTrue(
            'first document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )
        self.assertTrue(
            'second document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 11)

        test_documents = (Document.objects.first(), Document.objects.last())
        test_document_files = (
            test_documents[0].file_latest, test_documents[1].file_latest
        )
        test_document_versions = (
            test_documents[0].version_active, test_documents[1].version_active
        )
        test_document_version_pages = (
            test_document_versions[0].pages.first(),
            test_document_versions[1].pages.first(),
            test_document_versions[1].pages.last()
        )

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_files[0])
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_files[0])
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_versions[0])
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_versions[0])
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_pages[0])
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self.test_document_type)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_documents[1])
        self.assertEqual(events[5].verb, event_document_created.id)

        self.assertEqual(events[6].action_object, test_documents[1])
        self.assertEqual(events[6].actor, self._test_case_user)
        self.assertEqual(events[6].target, test_document_files[1])
        self.assertEqual(events[6].verb, event_document_file_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, self._test_case_user)
        self.assertEqual(events[7].target, test_document_files[1])
        self.assertEqual(events[7].verb, event_document_file_edited.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, self._test_case_user)
        self.assertEqual(events[8].target, test_document_versions[1])
        self.assertEqual(events[8].verb, event_document_version_created.id)

        self.assertEqual(events[9].action_object, test_document_versions[1])
        self.assertEqual(events[9].actor, self._test_case_user)
        self.assertEqual(events[9].target, test_document_version_pages[1])
        self.assertEqual(
            events[9].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[10].action_object, test_document_versions[1])
        self.assertEqual(events[10].actor, self._test_case_user)
        self.assertEqual(events[10].target, test_document_version_pages[2])
        self.assertEqual(
            events[10].verb, event_document_version_page_created.id
        )

    def test_upload_interactive_view_no_permission(self):
        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_interactive_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_interactive_view_with_document_type_access(self):
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_interactive_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_interactive_view_with_source_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_interactive_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_interactive_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_interactive_view()
        self.assertContains(
            response=response, text=self.test_source.label, status_code=200
        )

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_wizard_no_permission(self):
        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_wizard_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_wizard_with_document_type_access(self):
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_wizard_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_wizard_with_source_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_wizard_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_upload_wizard_with_full_access(self):
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_wizard_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Document.objects.count(), test_document_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )


class WebFormDocumentFileUploadViewTestCase(
    DocumentFileUploadViewTestMixin, WebFormSourceTestMixin,
    GenericDocumentViewTestCase
):
    def test_document_file_upload_view_no_permission(self):
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_post_view()

        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_post_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_view_with_source_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_post_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_post_view()
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.files.count(), file_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

    def test_trashed_document_file_upload_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self.test_document.delete()

        self._clear_events()

        response = self._request_document_file_upload_post_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.files.count(), file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_no_source_view_no_permission(self):
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_no_source_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_no_source_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_no_source_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_no_source_view_with_source_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_no_source_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_no_source_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_no_source_view()
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )

    def test_trashed_document_file_upload_no_source_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self.test_document.delete()

        self._clear_events()

        response = self._request_document_file_upload_no_source_view()
        self.assertEqual(response.status_code, 404)

        self.test_document.refresh_from_db()
        self.assertEqual(self.test_document.files.count(), file_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_upload_view_preserve_filename(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )

        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_post_view()
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.files.count(), file_count + 1
        )
        self.assertEqual(
            self.test_document.file_latest.filename,
            self.test_document_filename
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document_file)
        self.assertEqual(events[0].verb, event_document_file_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_edited.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document_version)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version_page)
        self.assertEqual(
            events[3].verb, event_document_version_page_created.id
        )


class WebFormDocumentCreateSourceSwitchLinkViewTestCase(
    WebFormSourceTestMixin, DocumentUploadWizardViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def test_source_switch_link_with_access(self):
        self._create_test_web_form_source()
        self._create_test_web_form_source()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_sources[1], permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_upload_interactive_view()

        self.assertNotContains(
            response=response, text=self._test_sources[0].label, status_code=200
        )
        self.assertContains(
            response=response, text=self._test_sources[1].label, status_code=200
        )

        self.assertEqual(Document.objects.count(), test_document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class WebFormDocumentFileUploadSourceSwitchLinkViewTestCase(
    DocumentFileUploadViewTestMixin, WebFormSourceTestMixin,
    GenericDocumentViewTestCase
):
    auto_create_test_source = False

    def test_document_file_upload_view_with_full_access(self):
        self._create_test_web_form_source()
        self._create_test_web_form_source()

        self.grant_access(
            obj=self.test_document, permission=permission_document_file_new
        )
        self.grant_access(
            obj=self._test_sources[1], permission=permission_document_file_new
        )
        file_count = self.test_document.files.count()

        self._clear_events()

        response = self._request_document_file_upload_get_view()
        self.assertNotContains(
            response=response, text=self._test_sources[0].label, status_code=200
        )
        self.assertContains(
            response=response, text=self._test_sources[1].label, status_code=200
        )

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.files.count(), file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.core.files import File

from mayan.apps.documents.models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import (
    TEST_COMPRESSED_DOCUMENT_PATH, TEST_SMALL_DOCUMENT_CHECKSUM,
    TEST_SMALL_DOCUMENT_PATH
)

from ..source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_ALWAYS

from .mixins.base_mixins import InteractiveSourceBackendTestMixin
from .mixins.web_form_source_mixins import WebFormSourceTestMixin


class WebFormSourceBackendTestCase(
    InteractiveSourceBackendTestMixin, WebFormSourceTestMixin,
    GenericDocumentTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def _process_test_document(self, test_file_path=TEST_SMALL_DOCUMENT_PATH):
        source_backend_instance = self.test_source.get_backend_instance()

        with open(file=test_file_path, mode='rb') as file_object:
            self.test_forms = {
                'document_form': self.test_document_form,
                'source_form': InteractiveSourceBackendTestMixin.MockSourceForm(
                    file=File(file=file_object)
                ),
            }

            source_backend_instance.process_documents(
                document_type=self.test_document_type, forms=self.test_forms,
                request=self.get_test_request()
            )

    def test_upload_simple_file(self):
        self._create_test_web_form_source()

        document_count = Document.objects.count()

        self._process_test_document()

        self.assertEqual(Document.objects.count(), document_count + 1)
        self.assertEqual(
            Document.objects.first().file_latest.checksum,
            TEST_SMALL_DOCUMENT_CHECKSUM
        )

    def test_upload_compressed_file(self):
        self._create_test_web_form_source(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ALWAYS}
        )

        document_count = Document.objects.count()

        self._process_test_document(
            test_file_path=TEST_COMPRESSED_DOCUMENT_PATH
        )

        self.assertEqual(Document.objects.count(), document_count + 2)

        self.assertTrue(
            'first document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )
        self.assertTrue(
            'second document.pdf' in Document.objects.values_list(
                'label', flat=True
            )
        )



from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_smart_link_created, event_smart_link_edited
from ..models import SmartLink
from ..permissions import (
    permission_smart_link_create, permission_smart_link_delete,
    permission_smart_link_edit, permission_smart_link_view
)

from .literals import TEST_SMART_LINK_LABEL, TEST_SMART_LINK_LABEL_EDITED
from .mixins import (
    DocumentTypeAddRemoveSmartLinkViewTestMixin,
    SmartLinkDocumentTypeViewTestMixin, SmartLinkTestMixin,
    SmartLinkViewTestMixin
)


class DocumentTypeAddRemoveSmartLinkViewTestCase(
    DocumentTypeAddRemoveSmartLinkViewTestMixin, SmartLinkTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_smart_link()

    def test_document_type_smart_link_add_remove_get_view_no_permission(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self._clear_events()

        response = self._request_test_document_type_smart_link_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_smart_link),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_add_remove_get_view_with_document_type_access(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_document_type),
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self.test_smart_link),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_add_remove_get_view_with_smart_link_access(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_smart_link),
            status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_add_remove_get_view_with_full_access(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_smart_link),
            status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_add_view_no_permission(self):
        self._clear_events()

        response = self._request_test_document_type_smart_link_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_smart_link not in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_add_view_with_document_type_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_smart_link not in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_add_view_with_smart_link_access(self):
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_smart_link not in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_add_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_smart_link in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)

    def test_document_type_smart_link_remove_view_no_permission(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self._clear_events()

        response = self._request_test_document_type_smart_link_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_smart_link in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_remove_view_with_document_type_access(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertTrue(
            self.test_smart_link in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_remove_view_with_smart_link_access(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertTrue(
            self.test_smart_link in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_type_smart_link_remove_view_with_full_access(self):
        self.test_document_type.smart_links.add(self.test_smart_link)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_document_type_smart_link_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertTrue(
            self.test_smart_link not in self.test_document_type.smart_links.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)


class SmartLinkViewTestCase(
    SmartLinkTestMixin, SmartLinkViewTestMixin, GenericViewTestCase
):
    def test_smart_link_create_view_no_permission(self):
        self._clear_events()

        response = self._request_test_smart_link_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(SmartLink.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_create_view_with_permission(self):
        self.grant_permission(permission=permission_smart_link_create)

        self._clear_events()

        response = self._request_test_smart_link_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(SmartLink.objects.count(), 1)
        self.assertEqual(
            SmartLink.objects.first().label, TEST_SMART_LINK_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_created.id)

    def test_smart_link_delete_view_no_permission(self):
        self._create_test_smart_link()

        self._clear_events()

        response = self._request_test_smart_link_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(SmartLink.objects.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_delete_view_with_access(self):
        self._create_test_smart_link()
        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_delete
        )

        self._clear_events()

        response = self._request_test_smart_link_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(SmartLink.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_edit_view_no_permission(self):
        self._create_test_smart_link()

        self._clear_events()

        response = self._request_test_smart_link_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_smart_link.refresh_from_db()
        self.assertEqual(self.test_smart_link.label, TEST_SMART_LINK_LABEL)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_edit_view_with_access(self):
        self._create_test_smart_link()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_smart_link.refresh_from_db()
        self.assertEqual(
            self.test_smart_link.label, TEST_SMART_LINK_LABEL_EDITED
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)

    def test_smart_link_list_view_no_permission(self):
        self._create_test_smart_link()

        self._clear_events()

        response = self._request_test_smart_link_list_view()
        self.assertNotContains(
            response=response, text=str(self.test_smart_link), status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_list_view_with_access(self):
        self._create_test_smart_link()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_view
        )

        self._clear_events()

        response = self._request_test_smart_link_list_view()
        self.assertContains(
            response=response, text=str(self.test_smart_link), status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class SmartLinkDocumentTypeViewTestCase(
    SmartLinkDocumentTypeViewTestMixin, SmartLinkTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_smart_link()

    def test_smart_link_document_type_add_remove_get_view_no_permission(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_smart_link),
            status_code=404
        )

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_add_remove_get_view_with_document_type_access(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_document_type),
            status_code=404
        )
        self.assertNotContains(
            response=response, text=str(self.test_smart_link),
            status_code=404
        )

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_add_remove_get_view_with_smart_link_access(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_remove_get_view()
        self.assertNotContains(
            response=response, text=str(self.test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_smart_link),
            status_code=200
        )

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_add_remove_get_view_with_full_access(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_smart_link,
            permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_remove_get_view()
        self.assertContains(
            response=response, text=str(self.test_document_type),
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self.test_smart_link),
            status_code=200
        )

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_add_view_no_permission(self):
        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_add_view_with_document_type_access(self):
        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_add_view_with_smart_link_access(self):
        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_add_view_with_full_access(self):
        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_add_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)

    def test_smart_link_document_type_remove_view_no_permission(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_remove_view_with_document_type_access(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type, permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_remove_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_remove_view_with_smart_link_access(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_remove_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_smart_link_document_type_remove_view_with_full_access(self):
        self.test_smart_link.document_types.add(self.test_document_type)

        test_smart_link_document_type_count = self.test_smart_link.document_types.count()

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )

        self._clear_events()

        response = self._request_test_smart_link_document_type_remove_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_smart_link.document_types.count(),
            test_smart_link_document_type_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_smart_link)
        self.assertEqual(events[0].verb, event_smart_link_edited.id)



import logging

from celery.backends.base import DisabledBackend

from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_tools
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.views.html_widgets import TwoStateWidget
from mayan.celery import app as celery_app

from .classes import CeleryQueue, Task
from .links import link_task_manager
from .literals import TEST_CELERY_RESULT_KEY, TEST_CELERY_RESULT_VALUE

logger = logging.getLogger(name=__name__)


class TaskManagerApp(MayanAppConfig):
    app_namespace = 'task_manager'
    app_url = 'task_manager'
    has_tests = True
    name = 'mayan.apps.task_manager'
    verbose_name = _('Task manager')

    def check_broker_connectivity(self):
        connection = celery_app.connection()

        logger.debug('Starting Celery broker connectivity test')
        try:
            connection.ensure_connection(
                interval_step=0, interval_max=0, interval_start=0, timeout=0.1
            )
        except Exception as exception:
            print(
                'Failed to connect to the Celery broker at {}; {}'.format(
                    connection.as_uri(), exception
                )
            )
            raise
        else:
            connection.release()

    def check_results_backend_connectivity(self):
        backend = celery_app.backend

        if not isinstance(backend, DisabledBackend):
            retry_policy = backend.retry_policy

            backend.retry_policy = {
                'max_retries': 0, 'interval_start': 0, 'interval_step': 1,
                'interval_max': 1
            }

            logger.debug('Starting Celery result backend connectivity test')
            try:
                backend.set(
                    key=TEST_CELERY_RESULT_KEY, value=TEST_CELERY_RESULT_VALUE
                )
            except Exception as exception:
                print(
                    'Failed to connect to the Celery result backend at {}; {}'.format(
                        backend.as_uri(), exception
                    )
                )
                raise
            else:
                backend.delete(key=TEST_CELERY_RESULT_KEY)
                backend.retry_policy = retry_policy

    def ready(self):
        super().ready()

        try:
            self.check_broker_connectivity()
        except Exception as exception:
            print(
                'Error checking Celery broker connectivity: {}'.format(exception)
            )
            exit(1)

        try:
            self.check_results_backend_connectivity()
        except Exception as exception:
            print(
                'Error checking Celery result backend connectivity: {}'.format(exception)
            )
            exit(1)

        CeleryQueue.load_modules()

        SourceColumn(
            attribute='label', is_identifier=True, label=_('Label'),
            source=CeleryQueue
        )
        SourceColumn(
            attribute='name', include_label=True, label=_('Name'),
            source=CeleryQueue
        )
        SourceColumn(
            attribute='default_queue', include_label=True,
            label=_('Default queue?'), source=CeleryQueue,
            widget=TwoStateWidget
        )
        SourceColumn(
            attribute='transient', include_label=True,
            label=_('Is transient?'), source=CeleryQueue,
            widget=TwoStateWidget
        )
        SourceColumn(
            attribute='task_type', include_label=True, label=_('Type'),
            source=Task
        )
        SourceColumn(
            attribute='get_time_started', include_label=True,
            label=_('Start time'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['hostname'],
            include_label=True, label=_('Host'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['args'],
            include_label=True, label=_('Arguments'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['kwargs'],
            include_label=True, label=_('Keyword arguments'), source=Task
        )
        SourceColumn(
            func=lambda context: context['object'].kwargs['worker_pid'],
            include_label=True, label=_('Worker process ID'), source=Task
        )

        menu_tools.bind_links(links=(link_task_manager,))



import logging

from django.contrib import messages
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.literals import DOCUMENT_FILE_ACTION_PAGES_NEW
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.permissions import permission_document_file_new
from mayan.apps.views.mixins import ExternalObjectViewMixin

from ..forms import NewDocumentFileForm
from ..models import Source

from .base import UploadBaseView

__all__ = ('DocumentFileUploadInteractiveView',)
logger = logging.getLogger(name=__name__)


class DocumentFileUploadInteractiveView(
    ExternalObjectViewMixin, UploadBaseView
):
    document_form = NewDocumentFileForm
    external_object_queryset = Document.valid.all()
    external_object_permission = permission_document_file_new
    external_object_pk_url_kwarg = 'document_id'
    object_permission = permission_document_file_new

    def dispatch(self, request, *args, **kwargs):
        self.subtemplates_list = []

        result = super().dispatch(request, *args, **kwargs)

        try:
            DocumentFile.execute_pre_create_hooks(
                kwargs={
                    'document': self.external_object,
                    'file_object': None,
                    'user': self.request.user
                }
            )
        except Exception as exception:
            messages.error(
                message=_(
                    'Unable to upload new files for document '
                    '"%(document)s". %(exception)s'
                ) % {'document': self.external_object, 'exception': exception},
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    viewname='documents:document_file_list',
                    kwargs={'document_id': self.external_object.pk}
                )
            )

        return result

    def forms_valid(self, forms):
        source_backend_instance = self.source.get_backend_instance()

        try:
            source_backend_instance.process_document_file(
                document=self.external_object, forms=forms,
                request=self.request
            )
        except Exception as exception:
            message = _(
                'Error executing document file upload task; '
                '%(exception)s'
            ) % {
                'exception': exception,
            }
            logger.critical(msg=message, exc_info=True)
            if self.request.is_ajax():
                return JsonResponse(
                    data={'error': force_text(s=message)}, status=500
                )
            else:
                messages.error(
                    message=message.replace('\n', ' '),
                    request=self.request
                )
                raise type(exception)(message)
        else:
            messages.success(
                message=_(
                    'New document file queued for upload and will be '
                    'available shortly.'
                ), request=self.request
            )

        return HttpResponseRedirect(
            redirect_to=reverse(
                viewname='documents:document_file_list', kwargs={
                    'document_id': self.external_object.pk
                }
            )
        )

    def get_active_tab_links(self):
        sources = AccessControlList.objects.restrict_queryset(
            permission=permission_document_file_new,
            queryset=Source.objects.interactive().filter(enabled=True),
            user=self.request.user
        )
        return [
            UploadBaseView.get_tab_link_for_source(
                source=source, document=self.external_object
            )
            for source in sources
        ]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                'form_action': '{}?{}'.format(
                    reverse(
                        viewname=self.request.resolver_match.view_name,
                        kwargs=self.request.resolver_match.kwargs
                    ), self.request.META['QUERY_STRING']
                ),
                'object': self.external_object,
                'title': _(
                    'Upload a new file for document "%(document)s" '
                    'from source: %(source)s'
                ) % {
                    'document': self.external_object,
                    'source': self.source.label
                },
                'submit_label': _('Submit')
            }
        )
        context.update(
            self.source.get_backend_instance().get_view_context(
                context=context, request=self.request
            )
        )

        return context

    def get_form_extra_kwargs__source_form(self, **kwargs):
        return {
            'source': self.source,
        }

    def get_initial__document_form(self):
        return {'action': DOCUMENT_FILE_ACTION_PAGES_NEW}



from pathlib import Path
from unittest.mock import patch

from mayan.apps.documents.models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import (
    TEST_SMALL_DOCUMENT_CHECKSUM, TEST_SMALL_DOCUMENT_PATH
)

from .mixins.base_mixins import InteractiveSourceBackendTestMixin
from .mixins.staging_folder_source_mixins import StagingFolderTestMixin


class StagingFolderSourceBackendTestCase(
    StagingFolderTestMixin, InteractiveSourceBackendTestMixin,
    GenericDocumentTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def _process_test_document(self, test_file_path=TEST_SMALL_DOCUMENT_PATH):
        source_backend_instance = self.test_source.get_backend_instance()

        self.test_forms = {
            'document_form': self.test_document_form,
            'source_form': InteractiveSourceBackendTestMixin.MockSourceForm(
                staging_folder_file_id=self.test_staging_folder_file.encoded_filename
            ),
        }

        source_backend_instance.process_documents(
            document_type=self.test_document_type, forms=self.test_forms,
            request=self.get_test_request()
        )

    def test_upload_simple_file(self):
        self._create_test_staging_folder()

        self._copy_test_staging_folder_document()

        document_count = Document.objects.count()

        self._process_test_document()

        self.assertEqual(Document.objects.count(), document_count + 1)
        self.assertEqual(
            Document.objects.first().file_latest.checksum,
            TEST_SMALL_DOCUMENT_CHECKSUM
        )

    @patch('mayan.apps.sources.source_backends.mixins.SourceBackendInteractiveMixin.callback')
    def test_super_class_callback(self, mocked_super):
        self._create_test_staging_folder()

        self._copy_test_staging_folder_document()

        document_count = Document.objects.count()

        self._process_test_document()

        self.assertEqual(Document.objects.count(), document_count + 1)
        self.assertEqual(
            Document.objects.first().file_latest.checksum,
            TEST_SMALL_DOCUMENT_CHECKSUM
        )

        self.assertTrue(mocked_super.called)

    def test_delete_after_upload(self):
        self._create_test_staging_folder(
            extra_data={'delete_after_upload': True}
        )

        self._copy_test_staging_folder_document()

        document_count = Document.objects.count()

        self._process_test_document()

        self.assertEqual(Document.objects.count(), document_count + 1)
        self.assertEqual(
            Document.objects.first().file_latest.checksum,
            TEST_SMALL_DOCUMENT_CHECKSUM
        )

        path = Path(self.test_source.get_backend_data()['folder_path'])

        self.assertEqual(sum(1 for x in path.glob('*') if x.is_file()), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(label=_('Appearance'), name='appearance')

permission_theme_create = namespace.add_permission(
    label=_('Create new themes'), name='theme_create'
)
permission_theme_delete = namespace.add_permission(
    label=_('Delete themes'), name='theme_delete'
)
permission_theme_edit = namespace.add_permission(
    label=_('Edit themes'), name='theme_edit'
)
permission_theme_view = namespace.add_permission(
    label=_('View existing themes'), name='theme_view'
)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('document_parsing', '0005_rename_documentversionparseerror_field'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='DocumentVersionParseError',
            new_name='DocumentFileParseError',
        )
    ]



from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('documents', '0069_auto_20201128_0247'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='RecentDocument',
            new_name='RecentlyAccessedDocument',
        ),
    ]



from rest_framework.mixins import CreateModelMixin

from .literals import DEFAULT_DYNAMIC_FIELD_SEPARATOR


class CreateOnlyFieldSerializerMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.is_create_view():
            # Remove the create only fields if the view is anything other
            # than a create view.
            self._excluded_fields.update(
                getattr(self.Meta, 'create_only_fields', ())
            )

    def is_create_view(self):
        request = self.context.get('request')
        view = self.context.get('view')

        if request and view:
            if isinstance(view, CreateModelMixin) and request.method.lower() == 'post':
                # This is a create view with a request to create an instance.
                return True

        return False


class DynamicFieldListSerializerMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._excluded_fields = set()

        # Handle top level fields.
        fields_exclude = self.context.get('fields_exclude', '').split(',')
        fields_only = self.context.get('fields_only', '').split(',')
        matched_exclude = []
        matched_only = []

        for field_exclude in fields_exclude:
            parts = field_exclude.split(DEFAULT_DYNAMIC_FIELD_SEPARATOR)
            matched_exclude.append(parts[-1])

        for field_only in fields_only:
            parts = field_only.split(DEFAULT_DYNAMIC_FIELD_SEPARATOR)
            if parts[0]:
                matched_only.append(parts[0])

        self.update_excluded_fields(
            fields_exclude=matched_exclude, fields_only=matched_only
        )

    def bind(self, field_name, parent):
        super().bind(field_name=field_name, parent=parent)

        # Handle fields from included serializers.

        # Context retrieval is repeated to ensure context is available after
        # __init__.
        fields_exclude = self.context.get('fields_exclude', '').split(',')
        fields_only = self.context.get('fields_only', '').split(',')
        matched_exclude = []
        matched_only = []
        path = self.get_path()
        path_level = len(path)

        for field_exclude in fields_exclude:
            parts = field_exclude.split(DEFAULT_DYNAMIC_FIELD_SEPARATOR)
            if path == parts[0:path_level]:
                local_field_name = parts[-1]
                matched_exclude.append(local_field_name)

        for field_only in fields_only:
            parts = field_only.split(DEFAULT_DYNAMIC_FIELD_SEPARATOR)
            if path == parts[0:path_level]:
                try:
                    local_field_name = parts[path_level]
                except IndexError:
                    pass
                else:
                    matched_only.append(local_field_name)

        self.update_excluded_fields(
            fields_exclude=matched_exclude, fields_only=matched_only
        )

    def get_fields(self):
        fields = super().get_fields()

        # Copy keys to avoid modifying them in the loop.
        field_names = list(fields.keys())

        for field in field_names:
            if field in self._excluded_fields:
                fields.pop(field)

        return fields

    def get_path(self):
        result = [self.field_name]

        parent = self.parent
        while True:
            if parent:
                if parent.field_name:
                    result.append(parent.field_name)

                parent = parent.parent
            else:
                break

        result.reverse()

        return result

    def update_excluded_fields(self, fields_exclude=None, fields_only=None):
        fields_exclude = set(fields_exclude or ())
        fields_only = set(fields_only or ())

        self._excluded_fields.update(fields_exclude)

        if fields_only:
            serializer_fields = set(super().get_fields())

            self._excluded_fields.update(
                serializer_fields - fields_only
            )



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='gevent', version_string='==21.8.0'
)
PythonDependency(
    module=__name__, name='gunicorn', version_string='==20.1.0'
)
PythonDependency(
    module=__name__, name='sentry-sdk', version_string='==1.4.1'
)
PythonDependency(
    module=__name__, name='whitenoise', version_string='==5.3.0'
)



from django.contrib import admin

from .models import MetadataType


@admin.register(MetadataType)
class MetadataTypeAdmin(admin.ModelAdmin):
    list_display = (
        'name', 'label', 'default', 'lookup', 'validation', 'parser'
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Locales'), name='locales')

event_user_locale_profile_edited = namespace.add_event_type(
    label=_('User locale profile edited'), name='user_locale_profile_edited'
)



from .. import *  # NOQA

AUTHENTICATION_BACKEND = 'mayan.apps.authentication.authentication_backends.AuthenticationBackendModelDjangoDefault'

CELERY_BROKER_URL = 'memory://'
CELERY_TASK_ALWAYS_EAGER = True
CELERY_TASK_EAGER_PROPAGATES = True

COMMON_HOME_VIEW_DASHBOARD_NAME = None

DOCUMENT_PARSING_AUTO_PARSING = False

FILE_METADATA_AUTO_PROCESS = False

INSTALLED_APPS = [
    cls for cls in INSTALLED_APPS if cls != 'whitenoise.runserver_nostatic'
]

LOGGING_LOG_FILE_PATH = '/tmp/mayan-errors.log'
LOGGING_LEVEL = 'WARNING'

# Remove middlewares not used for tests.
# Remove whitenoise from middlewares. Causes out of memory errors during test
# suit.
MIDDLEWARE = [
    cls for cls in MIDDLEWARE if cls not in [
        'common.middleware.error_logging.ErrorLoggingMiddleware',
        'django.middleware.security.SecurityMiddleware',
        'corsheaders.middleware.CorsMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
        'common.middleware.timezone.TimezoneMiddleware',
        'common.middleware.ajax_redirect.AjaxRedirect',
        'whitenoise.middleware.WhiteNoiseMiddleware'
    ]
]

OCR_AUTO_OCR = False

# User a simpler password hasher
PASSWORD_HASHERS = (
    'django.contrib.auth.hashers.MD5PasswordHasher',
)

SEARCH_BACKEND = 'mayan.apps.dynamic_search.tests.backends.TestSearchBackend'

STATICFILES_STORAGE = None

# Cache templates in memory
TEMPLATES[0]['OPTIONS']['loaders'] = (  # NOQA: F405
    (
        'django.template.loaders.cached.Loader', (
            'django.template.loaders.filesystem.Loader',
            'django.template.loaders.app_directories.Loader',
        )
    ),
)

TESTING = True  # Silence the error logger for non critical HTTP404 and PermissionDenied



from django.utils.encoding import force_text

from rest_framework.exceptions import ParseError

from mayan.apps.rest_api import generics

from .api_view_mixins import SearchModelAPIViewMixin
from .classes import SearchBackend, SearchModel
from .serializers import SearchModelSerializer
from .utils import get_match_all_value


class APISearchView(SearchModelAPIViewMixin, generics.ListAPIView):
    """
    get: Perform a search operation
    """
    def get_queryset(self):
        search_model = self.get_search_model()

        # Override serializer class just before producing the queryset of
        # search results.
        self.serializer_class = search_model.serializer

        if search_model.permission:
            self.mayan_object_permissions = {
                'GET': (search_model.permission,)
            }

        query_dict = self.request.GET.dict().copy()
        query_dict.update(self.request.POST.dict())

        try:
            queryset = SearchBackend.get_instance().search(
                search_model=search_model, query=query_dict,
                user=self.request.user
            )
        except Exception as exception:
            raise ParseError(force_text(s=exception))

        return queryset

    def get_serializer(self, *args, **kwargs):
        if self.get_search_model_name():
            return super().get_serializer(*args, **kwargs)
        else:
            return None


class APIAdvancedSearchView(SearchModelAPIViewMixin, generics.ListAPIView):
    """
    get: Perform an advanced search operation
    """
    def get_queryset(self):
        self.search_model = self.get_search_model()

        # Override serializer class just before producing the queryset of
        # search results.
        self.serializer_class = self.search_model.serializer

        if self.search_model.permission:
            self.mayan_object_permissions = {
                'GET': (self.search_model.permission,)
            }

        query_dict = self.request.GET.dict().copy()
        query_dict.update(self.request.POST.dict())

        global_and_search = get_match_all_value(
            value=query_dict.get('_match_all')
        )

        try:
            queryset = SearchBackend.get_instance().search(
                global_and_search=global_and_search, query=query_dict,
                search_model=self.search_model, user=self.request.user
            )
        except Exception as exception:
            raise ParseError(force_text(s=exception))

        return queryset

    def get_serializer(self, *args, **kwargs):
        if self.get_search_model_name():
            return super().get_serializer(*args, **kwargs)
        else:
            return None


class APISearchModelList(generics.ListAPIView):
    """
    get: Returns a list of all the available search models.
    """
    serializer_class = SearchModelSerializer

    def get_queryset(self):
        # This changes after the initial startup as search models are
        # automatically loaded.
        return SearchModel.all()



from mayan.apps.acls.models import AccessControlList
from mayan.apps.testing.tests.base import BaseTestCase

from ..models import EventSubscription, Notification, ObjectEventSubscription
from ..permissions import permission_events_view

from .mixins import NotificationTestMixin


class EventNotificationModelTestCase(NotificationTestMixin, BaseTestCase):
    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_local_test_user()
        self._create_local_test_object()

    def test_event_notification_single_user_no_permission(self):
        notification_count = Notification.objects.count()

        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )

        self._test_event_type.commit(target=self.test_object)

        self.assertEqual(Notification.objects.count(), notification_count)

    def test_event_notification_single_user_with_access(self):
        notification_count = Notification.objects.count()

        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )

        AccessControlList.objects.grant(
            obj=self.test_object, permission=permission_events_view,
            role=self.test_role
        )

        result = self._test_event_type.commit(target=self.test_object)

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_user)
        self.assertEqual(notification.action, result)

    def test_event_notification_multiple_users_with_user_0_access(self):
        self._create_local_test_user()

        notification_count = Notification.objects.count()

        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_object, permission=permission_events_view,
            role=self.test_roles[0]
        )

        result = self._test_event_type.commit(target=self.test_object)

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_users[0])
        self.assertEqual(notification.action, result)

    def test_event_notification_multiple_users_with_user_1_access(self):
        self._create_local_test_user()

        notification_count = Notification.objects.count()

        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_object, permission=permission_events_view,
            role=self.test_roles[1]
        )

        result = self._test_event_type.commit(target=self.test_object)

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_users[1])
        self.assertEqual(notification.action, result)

    def test_event_notification_multiple_users_with_all_user_access(self):
        self._create_local_test_user()

        notification_count = Notification.objects.count()

        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_object, permission=permission_events_view,
            role=self.test_roles[0]
        )
        AccessControlList.objects.grant(
            obj=self.test_object, permission=permission_events_view,
            role=self.test_roles[1]
        )

        result = self._test_event_type.commit(target=self.test_object)

        self.assertEqual(Notification.objects.count(), notification_count + 2)
        notifications = Notification.objects.all()
        user_pk_list = notifications.values_list('user__id', flat=True)
        self.assertTrue(self.test_users[0].pk in user_pk_list)
        self.assertEqual(notifications[0].action, result)
        self.assertTrue(self.test_users[1].pk in user_pk_list)
        self.assertEqual(notifications[1].action, result)


class ObjectEventNotificationModelTestCase(NotificationTestMixin, BaseTestCase):
    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_local_test_user()
        self._create_local_test_object()

    def test_object_notification_single_user_no_permission(self):
        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_object,
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )

        self._test_event_type.commit(target=self.test_object)

        self.assertEqual(Notification.objects.count(), notification_count)

    def test_object_notification_single_user_with_access(self):
        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_object,
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )

        AccessControlList.objects.grant(
            obj=self.test_object, permission=permission_events_view,
            role=self.test_role
        )

        result = self._test_event_type.commit(target=self.test_object)

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_user)
        self.assertEqual(notification.action, result)

    def test_object_notification_single_user_and_multiple_objects_with_object_0_access(self):
        self._create_local_test_object()
        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[1],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_role
        )

        result_0 = self._test_event_type.commit(target=self.test_objects[0])
        self._test_event_type.commit(target=self.test_objects[1])

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_user)
        self.assertEqual(notification.action, result_0)

    def test_object_notification_single_user_and_multiple_objects_with_object_0_target_access(self):
        self._create_local_test_object()
        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[1],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_role
        )

        result_0 = self._test_event_type.commit(
            target=self.test_objects[0], action_object=self.test_objects[1]
        )

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_user)
        self.assertEqual(notification.action, result_0)

    def test_object_notification_single_user_and_multiple_objects_with_object_1_action_object_access(self):
        self._create_local_test_object()
        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[1],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_user
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[1], permission=permission_events_view,
            role=self.test_role
        )

        result_0 = self._test_event_type.commit(
            target=self.test_objects[0], action_object=self.test_objects[1]
        )

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_user)
        self.assertEqual(notification.action, result_0)

    def test_object_notification_multiple_users_and_single_object_with_user_0_access(self):
        self._create_local_test_user()

        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_roles[0]
        )

        result_0 = self._test_event_type.commit(target=self.test_objects[0])

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_users[0])
        self.assertEqual(notification.action, result_0)

    def test_object_notification_multiple_users_and_single_object_with_user_1_access(self):
        self._create_local_test_user()

        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_roles[1]
        )

        result_0 = self._test_event_type.commit(target=self.test_objects[0])

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notification = Notification.objects.first()
        self.assertEqual(notification.user, self.test_users[1])
        self.assertEqual(notification.action, result_0)

    def test_object_notification_multiple_users_and_single_object_with_both_user_access(self):
        self._create_local_test_user()

        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_roles[0]
        )
        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_roles[1]
        )

        result_0 = self._test_event_type.commit(target=self.test_objects[0])

        self.assertEqual(Notification.objects.count(), notification_count + 2)
        notifications = Notification.objects.all()
        user_pk_list = notifications.values_list('user__id', flat=True)
        self.assertTrue(self.test_users[0].pk in user_pk_list)
        self.assertEqual(notifications[0].action, result_0)
        self.assertTrue(self.test_users[1].pk in user_pk_list)
        self.assertEqual(notifications[1].action, result_0)

    def test_object_notification_multiple_users_and_multiple_object_with_user_0_and_object_0_access(self):
        self._create_local_test_user()
        self._create_local_test_object()

        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[1],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[1],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_roles[0]
        )

        result_0 = self._test_event_type.commit(target=self.test_objects[0])
        self._test_event_type.commit(target=self.test_objects[1])

        self.assertEqual(Notification.objects.count(), notification_count + 1)
        notifications = Notification.objects.all()
        self.assertEqual(notifications[0].user, self.test_users[0])
        self.assertEqual(notifications[0].action, result_0)

    def test_object_notification_multiple_users_and_multiple_object_with_user_0_and_object_0_action_object_access(self):
        self._create_local_test_user()
        self._create_local_test_object()

        notification_count = Notification.objects.count()

        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[1],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[0]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[0],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )
        ObjectEventSubscription.objects.create(
            content_object=self.test_objects[1],
            stored_event_type=self._test_event_type.stored_event_type,
            user=self.test_users[1]
        )

        AccessControlList.objects.grant(
            obj=self.test_objects[0], permission=permission_events_view,
            role=self.test_roles[0]
        )

        result_0 = self._test_event_type.commit(
            action_object=self.test_objects[1], target=self.test_objects[0]
        )
        result_1 = self._test_event_type.commit(
            action_object=self.test_objects[0], target=self.test_objects[1]
        )

        self.assertEqual(Notification.objects.count(), notification_count + 2)
        notifications = Notification.objects.all()
        self.assertEqual(notifications[0].user, self.test_users[0])
        self.assertEqual(notifications[0].action, result_1)
        self.assertEqual(notifications[1].user, self.test_users[0])
        self.assertEqual(notifications[1].action, result_0)



from django.http import Http404

from ..classes import EventType
from ..literals import TEXT_UNKNOWN_EVENT_ID


class EventViewMixin:
    def get_event_type(self, id):
        try:
            return EventType.get(id=id)
        except KeyError:
            raise Http404(TEXT_UNKNOWN_EVENT_ID % id)



DEFAULT_CELERY_BROKER_LOGIN_METHOD = 'AMQPLAIN'
DEFAULT_CELERY_BROKER_URL = 'memory://'
DEFAULT_CELERY_BROKER_USE_SSL = None
DEFAULT_CELERY_RESULT_BACKEND = None

TEST_CELERY_RESULT_KEY = '__mayan_test_result'
TEST_CELERY_RESULT_VALUE = 'test'

WORKER_DEFAULT_CONCURRENCY = 0



from django.conf.urls import url

from .views.document_views import MailDocumentLinkView
from .views.document_file_views import (
    MailDocumentFileLinkView, MailDocumentFileAttachmentView
)
from .views.document_version_views import (
    MailDocumentVersionLinkView, MailDocumentVersionAttachmentView
)
from .views.mailing_profile_views import (
    UserMailerBackendSelectionView, UserMailingCreateView,
    UserMailingDeleteView, UserMailingEditView, UserMailerTestView,
    UserMailerListView
)


urlpatterns_document = [
    url(
        regex=r'^documents/(?P<document_id>\d+)/send/link/$',
        name='send_document_link_single',
        view=MailDocumentLinkView.as_view()
    ),
    url(
        regex=r'^documents/multiple/send/link/$',
        name='send_document_link_multiple',
        view=MailDocumentLinkView.as_view()
    )
]

urlpatterns_document_file = [
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/send/attachment/$',
        name='send_document_file_attachment_single',
        view=MailDocumentFileAttachmentView.as_view()
    ),
    url(
        regex=r'^documents/files/multiple/send/attachment/$',
        name='send_document_file_attachment_multiple',
        view=MailDocumentFileAttachmentView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/send/link/$',
        name='send_document_file_link_single',
        view=MailDocumentFileLinkView.as_view()
    ),
    url(
        regex=r'^documents/files/multiple/send/link/$',
        name='send_document_file_link_multiple',
        view=MailDocumentFileLinkView.as_view()
    )
]

urlpatterns_document_version = [
    url(
        regex=r'^documents/versions/(?P<document_version_id>\d+)/send/attachment/$',
        name='send_document_version_attachment_single',
        view=MailDocumentVersionAttachmentView.as_view()
    ),
    url(
        regex=r'^documents/versions/multiple/send/attachment/$',
        name='send_document_version_attachment_multiple',
        view=MailDocumentVersionAttachmentView.as_view()
    ),
    url(
        regex=r'^documents/versions/(?P<document_version_id>\d+)/send/link/$',
        name='send_document_version_link_single',
        view=MailDocumentVersionLinkView.as_view()
    ),
    url(
        regex=r'^documents/versions/multiple/send/link/$',
        name='send_document_version_link_multiple',
        view=MailDocumentVersionLinkView.as_view()
    )
]

urlpatterns_user_mailers = [
    url(
        regex=r'^user_mailers/backend/selection/$',
        name='user_mailer_backend_selection',
        view=UserMailerBackendSelectionView.as_view()
    ),
    url(
        regex=r'^user_mailers/(?P<class_path>[a-zA-Z0-9_.]+)/create/$',
        name='user_mailer_create', view=UserMailingCreateView.as_view()
    ),
    url(
        regex=r'^user_mailers/(?P<mailer_id>\d+)/delete/$',
        name='user_mailer_delete', view=UserMailingDeleteView.as_view()
    ),
    url(
        regex=r'^user_mailers/(?P<mailer_id>\d+)/edit/$',
        name='user_mailer_edit', view=UserMailingEditView.as_view()
    ),
    url(
        regex=r'^user_mailers/(?P<mailer_id>\d+)/test/$',
        name='user_mailer_test', view=UserMailerTestView.as_view()
    ),
    url(
        regex=r'^user_mailers/$', name='user_mailer_list',
        view=UserMailerListView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_document)
urlpatterns.extend(urlpatterns_document_file)
urlpatterns.extend(urlpatterns_document_version)
urlpatterns.extend(urlpatterns_user_mailers)



from mayan.apps.documents.tests.base import GenericDocumentTestCase


class ActionTestCase(GenericDocumentTestCase):
    def setUp(self):
        super().setUp()

        class MockWorkflowInstance:
            document = self.test_document

        class MockEntryLog:
            workflow_instance = MockWorkflowInstance()

        self.entry_log = MockEntryLog()



from django.apps import apps
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.classes import DocumentCreateWizardStep
from mayan.apps.views.http import URL

from .forms import CabinetListForm
from .models import Cabinet
from .permissions import permission_cabinet_add_document


class DocumentCreateWizardStepCabinets(DocumentCreateWizardStep):
    form_class = CabinetListForm
    label = _('Select cabinets')
    name = 'cabinet_selection'
    number = 3

    @classmethod
    def condition(cls, wizard):
        Cabinet = apps.get_model(
            app_label='cabinets', model_name='Cabinet'
        )
        return Cabinet.objects.exists()

    @classmethod
    def get_form_kwargs(self, wizard):
        return {
            'help_text': _('Cabinets to which the document will be added.'),
            'permission': permission_cabinet_add_document,
            'queryset': Cabinet.objects.all(),
            'user': wizard.request.user
        }

    @classmethod
    def done(cls, wizard):
        result = {}
        cleaned_data = wizard.get_cleaned_data_for_step(cls.name)
        if cleaned_data:
            result['cabinets'] = [
                force_text(s=cabinet.pk) for cabinet in cleaned_data['cabinets']
            ]

        return result

    @classmethod
    def step_post_upload_process(cls, document, query_string=None):
        Cabinet = apps.get_model(app_label='cabinets', model_name='Cabinet')
        cabinet_id_list = URL(query_string=query_string).args.getlist('cabinets')

        for cabinet in Cabinet.objects.filter(pk__in=cabinet_id_list):
            cabinet.document_add(document=document)


DocumentCreateWizardStep.register(DocumentCreateWizardStepCabinets)



from mayan.apps.dependencies.classes import (
    BinaryDependency, PythonDependency
)
from mayan.apps.dependencies.environments import environment_testing

from .literals import DEFAULT_FIREFOX_GECKODRIVER_PATH

BinaryDependency(
    environment=environment_testing, label='firefox-geckodriver',
    module=__name__, name='geckodriver',
    path=DEFAULT_FIREFOX_GECKODRIVER_PATH
)
PythonDependency(
    environment=environment_testing, module=__name__, name='coverage',
    version_string='==5.5'
)
PythonDependency(
    environment=environment_testing, module=__name__, name='coveralls',
    version_string='==3.2.0'
)
PythonDependency(
    environment=environment_testing, module=__name__,
    name='django-test-migrations', version_string='==1.1.0'
)
# Mock is set to production so that it is available in the Docker image
# and allows running the test suit in production.
PythonDependency(
    module=__name__, name='mock', version_string='==4.0.3'
)
PythonDependency(
    environment=environment_testing, module=__name__, name='selenium',
    version_string='==3.141.0'
)
PythonDependency(
    environment=environment_testing, module=__name__, name='tox',
    version_string='==3.27.0'
)
PythonDependency(
    environment=environment_testing, module=__name__, name='psutil',
    version_string='==5.8.0'
)



from django.shortcuts import get_object_or_404

from mayan.apps.acls.models import AccessControlList
from mayan.apps.converter.api_view_mixins import APIImageViewMixin
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_view
from mayan.apps.documents.serializers.document_type_serializers import DocumentTypeSerializer
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin
from mayan.apps.rest_api import generics

from ..models.workflow_models import Workflow
from ..permissions import (
    permission_workflow_template_create, permission_workflow_template_delete,
    permission_workflow_template_edit, permission_workflow_template_view
)
from ..serializers import (
    WorkflowTemplateDocumentTypeAddSerializer,
    WorkflowTemplateDocumentTypeRemoveSerializer, WorkflowTemplateSerializer,
    WorkflowTemplateStateActionSerializer, WorkflowTemplateStateSerializer,
    WorkflowTemplateTransitionSerializer, WorkflowTransitionFieldSerializer,
    WorkflowTemplateTransitionTriggerSerializer
)


class APIWorkflowTemplateDocumentTypeListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the document types attached to a workflow template.
    """
    external_object_class = Workflow
    external_object_pk_url_kwarg = 'workflow_template_id'
    mayan_external_object_permissions = {
        'GET': (permission_workflow_template_view,)
    }
    mayan_object_permissions = {
        'GET': (permission_document_type_view,),
    }
    serializer_class = DocumentTypeSerializer

    def get_queryset(self):
        """
        This view returns a list of document types that belong to a workflow template.
        """
        return self.external_object.document_types.all()


class APIWorkflowTemplateDocumentTypeAddView(generics.ObjectActionAPIView):
    """
    post: Add a document type to a workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permissions = {
        'POST': (permission_workflow_template_edit,)
    }
    serializer_class = WorkflowTemplateDocumentTypeAddSerializer
    queryset = Workflow.objects.all()

    def object_action(self, request, serializer):
        document_type = serializer.validated_data['document_type_id']
        self.object._event_actor = self.request.user
        self.object.document_types_add(
            queryset=DocumentType.objects.filter(pk=document_type.id)
        )


class APIWorkflowTemplateDocumentTypeRemoveView(generics.ObjectActionAPIView):
    """
    post: Remove a document type from a workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permissions = {
        'POST': (permission_workflow_template_edit,)
    }
    serializer_class = WorkflowTemplateDocumentTypeRemoveSerializer
    queryset = Workflow.objects.all()

    def object_action(self, request, serializer):
        document_type = serializer.validated_data['document_type_id']
        self.object._event_actor = self.request.user
        self.object.document_types_remove(
            queryset=DocumentType.objects.filter(pk=document_type.id)
        )


class APIWorkflowTemplateImageView(
    APIImageViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permissions = {
        'GET': (permission_workflow_template_view,),
    }
    queryset = Workflow.objects.all()


class APIWorkflowTemplateListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the workflow templates.
    post: Create a new workflow template.
    """
    mayan_object_permissions = {'GET': (permission_workflow_template_view,)}
    mayan_view_permissions = {'POST': (permission_workflow_template_create,)}
    ordering_fields = ('id', 'internal_name', 'label')
    queryset = Workflow.objects.all()
    serializer_class = WorkflowTemplateSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }


class APIWorkflowTemplateDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected workflow template.
    get: Return the details of the selected workflow template.
    patch: Edit the selected workflow template.
    put: Edit the selected workflow template.
    """
    lookup_url_kwarg = 'workflow_template_id'
    mayan_object_permissions = {
        'DELETE': (permission_workflow_template_delete,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,)
    }
    queryset = Workflow.objects.all()
    serializer_class = WorkflowTemplateSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }


# Workflow state views


class APIWorkflowTemplateStateListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the workflow template states.
    post: Create a new workflow template state.
    """
    ordering_fields = ('completion', 'id', 'initial', 'label')
    serializer_class = WorkflowTemplateStateSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'workflow': self.get_workflow_template()
        }

    def get_queryset(self):
        return self.get_workflow_template().states.all()

    def get_workflow_template(self):
        if self.request.method == 'GET':
            permission_required = permission_workflow_template_view
        else:
            permission_required = permission_workflow_template_edit

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_required, queryset=Workflow.objects.all(),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['workflow_template_id']
        )


class APIWorkflowTemplateStateView(
    ExternalObjectAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template state.
    get: Return the details of the selected workflow template state.
    patch: Edit the selected workflow template state.
    put: Edit the selected workflow template state.
    """
    external_object_class = Workflow
    external_object_pk_url_kwarg = 'workflow_template_id'
    mayan_external_object_permissions = {
        'DELETE': (permission_workflow_template_edit,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,),
    }
    lookup_url_kwarg = 'workflow_template_state_id'
    serializer_class = WorkflowTemplateStateSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }

    def get_queryset(self):
        return self.external_object.states.all()


class APIWorkflowTemplateStateActionListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the workflow template state actions.
    post: Create a new workflow template state action.
    """
    ordering_fields = ('label', 'enabled', 'id')
    serializer_class = WorkflowTemplateStateActionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'state': self.get_workflow_template_state()
        }

    def get_queryset(self):
        return self.get_workflow_template_state().actions.all()

    def get_workflow_template_state(self):
        if self.request.method == 'GET':
            permission_required = permission_workflow_template_view
        else:
            permission_required = permission_workflow_template_edit

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_required,
            queryset=self.get_workflow_template().states.all(),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['workflow_template_state_id']
        )

    def get_workflow_template(self):
        return get_object_or_404(
            klass=Workflow.objects.all(),
            pk=self.kwargs['workflow_template_id']
        )


class APIWorkflowTemplateStateActionDetailView(
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template state action.
    get: Return the details of the selected workflow template state action.
    patch: Edit the selected workflow template state action.
    put: Edit the selected workflow template state action.
    """
    mayan_object_permissions = {
        'DELETE': (permission_workflow_template_edit,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,),
    }
    lookup_url_kwarg = 'workflow_template_state_action_id'
    serializer_class = WorkflowTemplateStateActionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }

    def get_queryset(self):
        return self.get_workflow_template_state().actions.all()

    def get_workflow_template_state(self):
        if self.request.method == 'GET':
            permission_required = permission_workflow_template_view
        else:
            permission_required = permission_workflow_template_edit

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_required,
            queryset=self.get_workflow_template().states.all(),
            user=self.request.user
        )
        return get_object_or_404(
            klass=queryset, pk=self.kwargs['workflow_template_state_id']
        )

    def get_workflow_template(self):
        return get_object_or_404(
            klass=Workflow.objects.all(),
            pk=self.kwargs['workflow_template_id']
        )


# Workflow transition views


class APIWorkflowTemplateTransitionListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template transitions.
    post: Create a new workflow template transition.
    """
    external_object_class = Workflow
    external_object_pk_url_kwarg = 'workflow_template_id'
    mayan_external_object_permissions = {
        'GET': (permission_workflow_template_view,),
        'POST': (permission_workflow_template_edit,),
    }
    ordering_fields = ('destination_state', 'id', 'label', 'origin_state')
    serializer_class = WorkflowTemplateTransitionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'workflow': self.external_object,
        }

    def get_queryset(self):
        return self.external_object.transitions.all()


class APIWorkflowTemplateTransitionView(
    ExternalObjectAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template transition.
    get: Return the details of the selected workflow template transition.
    patch: Edit the selected workflow template transition.
    put: Edit the selected workflow template transition.
    """
    external_object_class = Workflow
    external_object_pk_url_kwarg = 'workflow_template_id'
    mayan_external_object_permissions = {
        'DELETE': (permission_workflow_template_edit,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,),
    }
    lookup_url_kwarg = 'workflow_template_transition_id'
    serializer_class = WorkflowTemplateTransitionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'workflow': self.external_object,
        }

    def get_queryset(self):
        return self.external_object.transitions.all()


# Workflow template transition fields


class APIWorkflowTemplateTransitionFieldListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the workflow template transition fields.
    post: Create a new workflow template transition field.
    """
    ordering_fields = ('id', 'label', 'name', 'required', 'widget_kwargs')
    serializer_class = WorkflowTransitionFieldSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'transition': self.get_workflow_template_transition(),
        }

    def get_queryset(self):
        return self.get_workflow_template_transition().fields.all()

    def get_workflow_template(self):
        if self.request.method == 'GET':
            permission_required = permission_workflow_template_view
        else:
            permission_required = permission_workflow_template_edit

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_required, queryset=Workflow.objects.all(),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['workflow_template_id']
        )

    def get_workflow_template_transition(self):
        return get_object_or_404(
            klass=self.get_workflow_template().transitions,
            pk=self.kwargs['workflow_template_transition_id']
        )


class APIWorkflowTemplateTransitionFieldDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected workflow template transition field.
    get: Return the details of the selected workflow template transition field.
    patch: Edit the selected workflow template transition field.
    put: Edit the selected workflow template transition field.
    """
    lookup_url_kwarg = 'workflow_template_transition_field_id'
    serializer_class = WorkflowTransitionFieldSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }

    def get_queryset(self):
        return self.get_workflow_template_transition().fields.all()

    def get_workflow_template(self):
        if self.request.method == 'GET':
            permission_required = permission_workflow_template_view
        else:
            permission_required = permission_workflow_template_edit

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_required, queryset=Workflow.objects.all(),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['workflow_template_id']
        )

    def get_workflow_template_transition(self):
        return get_object_or_404(
            klass=self.get_workflow_template().transitions,
            pk=self.kwargs['workflow_template_transition_id']
        )


# Workflow template transition triggers


class APIWorkflowTemplateTransitionTriggerListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the workflow template transition triggers.
    post: Create a new workflow template transition trigger.
    """
    ordering_fields = ('event_type', 'id')
    serializer_class = WorkflowTemplateTransitionTriggerSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'transition': self.get_workflow_template_transition()
        }

    def get_queryset(self):
        return self.get_workflow_template_transition().trigger_events.all()

    def get_workflow_template_transition(self):
        if self.request.method == 'GET':
            permission_required = permission_workflow_template_view
        else:
            permission_required = permission_workflow_template_edit

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_required,
            queryset=self.get_workflow_template().transitions.all(),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs['workflow_template_transition_id']
        )

    def get_workflow_template(self):
        return get_object_or_404(
            klass=Workflow.objects.all(),
            pk=self.kwargs['workflow_template_id']
        )


class APIWorkflowTemplateTransitionTriggerDetailView(
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template transition trigger.
    get: Return the details of the selected workflow template transition trigger.
    patch: Edit the selected workflow template transition trigger.
    put: Edit the selected workflow template transition trigger.
    """
    mayan_object_permissions = {
        'DELETE': (permission_workflow_template_edit,),
        'GET': (permission_workflow_template_view,),
        'PATCH': (permission_workflow_template_edit,),
        'PUT': (permission_workflow_template_edit,)
    }
    lookup_url_kwarg = 'workflow_template_transition_trigger_id'
    serializer_class = WorkflowTemplateTransitionTriggerSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }

    def get_queryset(self):
        return self.get_workflow_template_transition().trigger_events.all()

    def get_workflow_template_transition(self):
        if self.request.method == 'GET':
            permission_required = permission_workflow_template_view
        else:
            permission_required = permission_workflow_template_edit

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_required,
            queryset=self.get_workflow_template().transitions.all(),
            user=self.request.user
        )
        return get_object_or_404(
            klass=queryset, pk=self.kwargs['workflow_template_transition_id']
        )

    def get_workflow_template(self):
        return get_object_or_404(
            klass=Workflow.objects.all(),
            pk=self.kwargs['workflow_template_id']
        )



from mayan.apps.documents.models import Document
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .permissions import (
    permission_document_comment_create, permission_document_comment_delete,
    permission_document_comment_edit, permission_document_comment_view
)
from .serializers import CommentSerializer


class APICommentListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the document comments.
    post: Create a new document comment.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {
        'GET': (permission_document_comment_view,),
        'POST': (permission_document_comment_create,)
    }
    ordering_fields = ('id', 'submit_date')
    serializer_class = CommentSerializer

    def get_queryset(self):
        return self.external_object.comments.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'user': self.request.user,
            'document': self.external_object
        }


class APICommentView(
    ExternalObjectAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected document comment.
    get: Returns the details of the selected document comment.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {
        'DELETE': (permission_document_comment_delete,),
        'GET': (permission_document_comment_view,),
        'PATCH': (permission_document_comment_edit,),
        'PUT': (permission_document_comment_edit,),
    }
    lookup_url_kwarg = 'comment_id'
    serializer_class = CommentSerializer

    def get_queryset(self):
        return self.external_object.comments.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'document': self.external_object
        }



from io import BytesIO

from PIL import Image

from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from mayan.apps.converter.layers import layer_saved_transformations
from mayan.apps.converter.tests.literals import TEST_TRANSFORMATION_DOCUMENT_PATH
from mayan.apps.converter.transformations import TransformationRotate270
from mayan.apps.documents.tests.mixins.document_file_mixins import DocumentFilePageAPIViewTestMixin
from mayan.apps.documents.permissions import permission_document_file_view
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin

from ..layers import layer_redactions
from ..transformations import TransformationRedactionPercent
from ..permissions import permission_redaction_exclude


class LayerMaximumOrderAPIViewTestCase(
    DocumentFilePageAPIViewTestMixin, DocumentTestMixin, BaseAPITestCase
):
    test_document_path = TEST_TRANSFORMATION_DOCUMENT_PATH

    def test_redaction_maximum_layer_order_exclude_transformation(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_view
        )

        layer_saved_transformations.add_transformation_to(
            obj=self.test_document_file_page,
            transformation_class=TransformationRotate270
        )

        self._clear_events()

        response = self._request_test_document_file_page_image_api_view(
            maximum_layer_order=layer_redactions.order
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        image_buffer = BytesIO(b''.join(response.streaming_content))
        image = Image.open(fp=image_buffer)

        self.assertEqual(image.getpixel(xy=(0, 0)), (254, 0, 0))

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_image_request_with_layer_below_redaction_no_permission(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_view
        )

        layer_redactions.add_transformation_to(
            obj=self.test_document_file_page,
            transformation_class=TransformationRedactionPercent(
                left=0, top=0, right=95, bottom=95
            )
        )

        self._clear_events()

        response = self._request_test_document_file_page_image_api_view(
            maximum_layer_order=layer_redactions.order - 1
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        image_buffer = BytesIO(b''.join(response.streaming_content))
        image = Image.open(fp=image_buffer)

        self.assertEqual(image.getpixel(xy=(0, 0)), (0, 0, 0))

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_image_request_with_layer_below_redaction_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_view
        )
        self.grant_access(
            obj=self.test_document, permission=permission_redaction_exclude
        )

        layer_redactions.add_transformation_to(
            obj=self.test_document_file_page,
            transformation_class=TransformationRedactionPercent(
                left=0, top=0, right=95, bottom=95
            )
        )

        self._clear_events()

        response = self._request_test_document_file_page_image_api_view(
            maximum_layer_order=layer_redactions.order - 1
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        image_buffer = BytesIO(b''.join(response.streaming_content))
        image = Image.open(fp=image_buffer)

        self.assertEqual(image.getpixel(xy=(0, 0)), (254, 0, 0))

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



"""
Authentication adapters for Allauth
"""
try:
    from allauth.account.adapter import DefaultAccountAdapter
except ImportError:
    print('ERROR: This authentication adapter requires django-allauth.')
    raise

from django.contrib import messages
from django.utils.translation import ugettext_lazy as _

from .literals import ADMIN_EMAIL_ADDRESSES


class AutoadminAccountAdapter(DefaultAccountAdapter):
    """
    Allauth account adapter that enables automatic grant of admin permissions
    to users signing up having their email address listed in the ``ADMINS``
    Django settings.  Django settings needed to activate this feature:

        INSTALLED_APPS = [
            'django.contrib.auth',
            'django.contrib.sites',
            'allauth',
            'allauth.account',
            'allauth.socialaccount',
        ]

        ACCOUNT_ADAPTER = 'autoadmin.auth.allauth.AutoadminAccountAdapter'

    See also:
    - http://django-allauth.readthedocs.io/en/latest/configuration.html
    - http://django-allauth.readthedocs.io/en/latest/advanced.html#admin
    """

    def confirm_email(self, request, email_address):
        """
        Give superuser privileges automagically if the email address of a
        user confirming their email is listed in ``settings.ADMINS``.
        """
        super().confirm_email(request=request, email_address=email_address)

        if email_address.email in ADMIN_EMAIL_ADDRESSES:
            user = email_address.user
            user.is_staff = user.is_superuser = True
            user.save()

            messages.info(
                request=request, message=_(
                    'Welcome Admin! You have been given superuser '
                    'privileges. Use them with caution.'
                )
            )



from mayan.apps.appearance.classes import Icon
from mayan.apps.converter.icons import icon_transformations

icon_document_type = Icon(
    driver_name='fontawesome-layers', data=[
        {'class': 'fas fa-circle', 'transform': 'shrink-12 up-2'},
        {'class': 'fas fa-cog', 'transform': 'shrink-6 up-2', 'mask': 'fas fa-torah'}
    ]
)

icon_menu_documents = Icon(driver_name='fontawesome', symbol='book')

icon_dashboard_document_types = icon_document_type
icon_dashboard_documents_in_trash = Icon(
    driver_name='fontawesome', symbol='trash-alt'
)
icon_dashboard_pages_per_month = Icon(
    driver_name='fontawesome', symbol='copy'
)
icon_dashboard_new_documents_this_month = Icon(
    driver_name='fontawesome', symbol='calendar'
)
icon_dashboard_total_document = Icon(
    driver_name='fontawesome', symbol='book'
)
icon_document_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_document = Icon(driver_name='fontawesome', symbol='book')
icon_document_list = icon_document
icon_document_file_page_count_update = Icon(
    driver_name='fontawesome', symbol='copy'
)
icon_document_preview = Icon(driver_name='fontawesome', symbol='eye')
icon_document_properties = Icon(driver_name='fontawesome', symbol='info')
icon_document_trash_send = Icon(
    driver_name='fontawesome', symbol='trash-alt'
)


icon_document_type_change = icon_document_type

icon_document_image_loading = Icon(
    driver_name='fontawesomecss', css_classes='far fa-clock fa-2x'
)
icon_document_list = Icon(driver_name='fontawesome', symbol='book')
icon_document_list_deleted = Icon(
    driver_name='fontawesome', symbol='trash-alt'
)
icon_document_recently_accessed_list = Icon(
    driver_name='fontawesome', symbol='clock'
)
icon_document_return = Icon(
    driver_name='fontawesome-dual', primary_symbol='book',
    secondary_symbol='chevron-left'
)

# Favorites

icon_favorite_document_add = Icon(
    driver_name='fontawesome-dual', primary_symbol='star',
    secondary_symbol='plus'
)
icon_favorite_document_list = Icon(driver_name='fontawesome', symbol='star')
icon_favorite_document_remove = Icon(
    driver_name='fontawesome-dual', primary_symbol='star',
    secondary_symbol='minus'
)

# Document files

icon_document_file_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_document_file_download_quick = Icon(
    driver_name='fontawesome', symbol='download'
)
icon_document_file_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_document_file_list = Icon(
    driver_name='fontawesome', symbol='hdd'
)
icon_document_file_preview = Icon(
    driver_name='fontawesome', symbol='eye'
)
icon_document_file_print = Icon(
    driver_name='fontawesome', symbol='print'
)
icon_document_file_properties = Icon(
    driver_name='fontawesome', symbol='info'
)
icon_document_file_return_to_document = icon_document_return
icon_document_file_return_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='hdd',
    secondary_symbol='chevron-left'
)
icon_document_file_transformations_clear = Icon(
    driver_name='fontawesome-dual',
    primary_symbol=icon_transformations.kwargs['symbol'],
    secondary_symbol='times'
)
icon_document_file_transformations_clone = Icon(
    driver_name='fontawesome-dual',
    primary_symbol=icon_transformations.kwargs['symbol'],
    secondary_symbol='arrow-right'
)

# Document file pages

icon_document_file_page_list = Icon(driver_name='fontawesome', symbol='copy')
icon_document_file_page_navigation_first = Icon(
    driver_name='fontawesome', symbol='step-backward'
)
icon_document_file_page_navigation_last = Icon(
    driver_name='fontawesome', symbol='step-forward'
)
icon_document_file_page_navigation_next = Icon(
    driver_name='fontawesome', symbol='arrow-right'
)
icon_document_file_page_navigation_previous = Icon(
    driver_name='fontawesome', symbol='arrow-left'
)
icon_document_file_page_return_to_document = icon_document_return
icon_document_file_page_return_to_document_file = Icon(
    driver_name='fontawesome-dual', primary_symbol='hdd',
    secondary_symbol='chevron-left'
)
icon_document_file_page_return_to_document_file_page_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='copy',
    secondary_symbol='chevron-left'
)
icon_document_file_page_rotate_left = Icon(
    driver_name='fontawesome', symbol='undo'
)
icon_document_file_page_rotate_right = Icon(
    driver_name='fontawesome', symbol='redo'
)
icon_document_file_page_view = Icon(
    driver_name='fontawesome', symbol='image'
)
icon_document_file_page_view_reset = Icon(
    driver_name='fontawesome', symbol='sync'
)
icon_document_file_page_zoom_in = Icon(
    driver_name='fontawesome', symbol='search-plus'
)
icon_document_file_page_zoom_out = Icon(
    driver_name='fontawesome', symbol='search-minus'
)

# Document types

icon_document_type_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='book',
    secondary_symbol='plus'
)
icon_document_type_delete = Icon(driver_name='fontawesome', symbol='times')
icon_document_type_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_document_type_setup = icon_document_type


icon_document_type_list = icon_document_type

icon_document_type_filename = Icon(
    driver_name='fontawesome', symbol='keyboard'
)
icon_document_type_filename_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='keyboard',
    secondary_symbol='plus'
)
icon_document_type_filename_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_document_type_filename_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_document_type_filename_list = Icon(
    driver_name='fontawesome', symbol='keyboard'
)

icon_document_type_filename_generator = Icon(
    driver_name='fontawesome-layers', data=[
        {'class': 'far fa-file'},
        {'class': 'fas fa-cog', 'transform': 'shrink-8 down-2'}
    ]
)

icon_document_type_policies = Icon(driver_name='fontawesome', symbol='times')
icon_document_type_setup = icon_document_type

# Document version

icon_document_version_active = Icon(
    driver_name='fontawesome', symbol='check'
)
icon_document_version_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='code-branch',
    secondary_symbol='plus'
)
icon_document_version_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_document_version_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_document_version_export = Icon(
    driver_name='fontawesome', symbol='file-export'
)
icon_document_version_list = Icon(
    driver_name='fontawesome', symbol='code-branch'
)
icon_document_version_return_document = icon_document_return
icon_document_version_return_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='code-branch',
    secondary_symbol='chevron-left'
)
icon_document_version_preview = Icon(
    driver_name='fontawesome', symbol='eye'
)
icon_document_version_print = Icon(
    driver_name='fontawesome', symbol='print'
)
icon_document_version_transformations_clear = Icon(
    driver_name='fontawesome-dual',
    primary_symbol=icon_transformations.kwargs['symbol'],
    secondary_symbol='times'
)
icon_document_version_transformations_clone = Icon(
    driver_name='fontawesome-dual',
    primary_symbol=icon_transformations.kwargs['symbol'],
    secondary_symbol='arrow-right'
)

# Document version pages

icon_document_version_page_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_document_version_page_return_to_document = icon_document_return
icon_document_version_page_return_to_document_version = Icon(
    driver_name='fontawesome-dual', primary_symbol='code-branch',
    secondary_symbol='chevron-left'
)
icon_document_version_page_return_to_document_version_page_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='copy',
    secondary_symbol='chevron-left'
)
icon_document_version_page_list = Icon(
    driver_name='fontawesome', symbol='copy'
)
icon_document_version_page_list_append = Icon(
    driver_name='fontawesome', symbol='layer-group'
)
icon_document_version_page_list_remap = Icon(
    driver_name='fontawesome', symbol='project-diagram'
)
icon_document_version_page_list_reset = Icon(
    driver_name='fontawesome', symbol='copy'
)
icon_document_version_page_navigation_first = Icon(
    driver_name='fontawesome', symbol='step-backward'
)
icon_document_version_page_navigation_last = Icon(
    driver_name='fontawesome', symbol='step-forward'
)
icon_document_version_page_navigation_next = Icon(
    driver_name='fontawesome', symbol='arrow-right'
)
icon_document_version_page_navigation_previous = Icon(
    driver_name='fontawesome', symbol='arrow-left'
)
icon_document_version_page_rotate_left = Icon(
    driver_name='fontawesome', symbol='undo'
)
icon_document_version_page_rotate_right = Icon(
    driver_name='fontawesome', symbol='redo'
)
icon_document_version_page_view = Icon(
    driver_name='fontawesome', symbol='image'
)
icon_document_version_page_view_reset = Icon(
    driver_name='fontawesome', symbol='sync'
)
icon_document_version_page_zoom_in = Icon(
    driver_name='fontawesome', symbol='search-plus'
)
icon_document_version_page_zoom_out = Icon(
    driver_name='fontawesome', symbol='search-minus'
)

# Recently created

icon_document_recently_created_list = Icon(
    driver_name='fontawesome', symbol='asterisk'
)

# Trashed documents

icon_trash_can_empty = Icon(
    driver_name='fontawesome-dual-classes', primary_class='fas fa-trash-alt',
    secondary_class='fas fa-minus'
)
icon_trashed_document_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_trashed_document_list = Icon(
    driver_name='fontawesome', symbol='trash-alt'
)
icon_trashed_document_multiple_delete = Icon(
    driver_name='fontawesome', symbol='trash-alt'
)
icon_trashed_document_multiple_restore = Icon(
    driver_name='fontawesome', symbol='recycle'
)
icon_trashed_document_restore = Icon(
    driver_name='fontawesome', symbol='recycle'
)



import logging

from django.apps import apps
from django.db.models.signals import post_save
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelFieldRelated, ModelProperty
from mayan.apps.common.menus import (
    menu_list_facet, menu_multi_item, menu_secondary, menu_tools
)
from mayan.apps.documents.signals import signal_post_document_file_upload
from mayan.apps.events.classes import ModelEventType
from mayan.apps.navigation.classes import SourceColumn

from .events import (
    event_parsing_document_file_content_deleted,
    event_parsing_document_file_submitted,
    event_parsing_document_file_finished
)
from .handlers import (
    handler_index_document, handler_initialize_new_parsing_settings,
    handler_parse_document_file
)
from .links import (
    link_document_file_content, link_document_file_content_delete_single,
    link_document_file_content_delete_multiple, link_document_file_page_content,
    link_document_file_content_download, link_document_file_parsing_errors_list,
    link_document_file_metadata_submit_multiple, link_document_file_metadata_submit_single,
    link_document_type_parsing_settings, link_document_type_submit,
    link_error_list
)
from .methods import (
    method_document_content, method_document_file_content,
    method_document_parsing_submit, method_document_file_parsing_submit
)
from .permissions import (
    permission_document_file_content_view, permission_document_type_parsing_setup,
    permission_document_file_parse
)
from .signals import signal_post_document_file_parsing

logger = logging.getLogger(name=__name__)


class DocumentParsingApp(MayanAppConfig):
    app_namespace = 'document_parsing'
    app_url = 'parsing'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.document_parsing'
    verbose_name = _('Document parsing')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )
        DocumentFileParseError = self.get_model(
            model_name='DocumentFileParseError'
        )
        DocumentFilePage = apps.get_model(
            app_label='documents', model_name='DocumentFilePage'
        )
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        DocumentTypeSettings = self.get_model(
            model_name='DocumentTypeSettings'
        )

        Document.add_to_class(
            name='content', value=method_document_content
        )
        Document.add_to_class(
            name='submit_for_parsing', value=method_document_parsing_submit
        )
        DocumentFile.add_to_class(
            name='content', value=method_document_file_content
        )
        DocumentFile.add_to_class(
            name='submit_for_parsing',
            value=method_document_file_parsing_submit
        )

        ModelEventType.register(
            model=Document, event_types=(
                event_parsing_document_file_content_deleted,
                event_parsing_document_file_submitted,
                event_parsing_document_file_finished
            )
        )
        ModelEventType.register(
            model=DocumentFile, event_types=(
                event_parsing_document_file_content_deleted,
                event_parsing_document_file_submitted,
                event_parsing_document_file_finished
            )
        )

        ModelFieldRelated(
            model=Document, name='files__file_pages__content__content'
        )

        ModelProperty(
            description=_(
                'A generator returning the document file\'s pages parsed content.'
            ), label=_('Content'), model=Document,
            name='content'
        )

        ModelPermission.register(
            model=DocumentFile, permissions=(
                permission_document_file_content_view,
                permission_document_file_parse
            )
        )
        ModelPermission.register(
            model=DocumentType, permissions=(
                permission_document_type_parsing_setup,
            )
        )
        ModelPermission.register_inheritance(
            model=DocumentTypeSettings, related='document_type',
        )

        SourceColumn(
            attribute='document_file__document',
            is_attribute_absolute_url=True, is_identifier=True,
            is_sortable=True, source=DocumentFileParseError
        )
        SourceColumn(
            attribute='datetime_submitted', is_sortable=True,
            source=DocumentFileParseError
        )
        SourceColumn(
            source=DocumentFileParseError, label=_('Result'),
            attribute='result'
        )

        menu_list_facet.bind_links(
            links=(link_document_file_content,), sources=(DocumentFile,)
        )
        menu_list_facet.bind_links(
            links=(link_document_file_page_content,),
            sources=(DocumentFilePage,)
        )
        menu_list_facet.bind_links(
            links=(link_document_type_parsing_settings,),
            sources=(DocumentType,)
        )
        menu_multi_item.bind_links(
            links=(
                link_document_file_content_delete_multiple,
                link_document_file_metadata_submit_multiple,
            ), sources=(DocumentFile,)
        )
        menu_secondary.bind_links(
            links=(
                link_document_file_content_delete_single,
                link_document_file_content_download,
                link_document_file_parsing_errors_list,
                link_document_file_metadata_submit_single
            ),
            sources=(
                'document_parsing:document_file_content_view',
                'document_parsing:document_file_content_delete_single',
                'document_parsing:document_file_content_download',
                'document_parsing:document_file_parsing_error_list',
                'document_parsing:document_file_submit_single',
            )
        )
        menu_tools.bind_links(
            links=(
                link_document_type_submit, link_error_list,
            )
        )

        post_save.connect(
            dispatch_uid='document_parsing_handler_initialize_new_parsing_settings',
            receiver=handler_initialize_new_parsing_settings,
            sender=DocumentType
        )
        signal_post_document_file_parsing.connect(
            dispatch_uid='document_parsing_handler_index_document',
            receiver=handler_index_document,
            sender=DocumentFile
        )
        signal_post_document_file_upload.connect(
            dispatch_uid='document_parsing_handler_parse_document_file',
            receiver=handler_parse_document_file,
            sender=DocumentFile
        )



from django.template import Library

from ..classes import Menu, SourceColumn

register = Library()


def _navigation_resolve_menu(context, name, source=None, sort_results=None):
    result = []

    menu = Menu.get(name)
    link_groups = menu.resolve(
        context=context, source=source, sort_results=sort_results
    )

    if link_groups:
        result.append(
            {
                'link_groups': link_groups, 'menu': menu
            }
        )

    return result


@register.simple_tag(takes_context=True)
def navigation_get_sort_field_querystring(context, column):
    return column.get_sort_field_querystring(context=context)


@register.simple_tag
def navigation_get_source_columns(
    source, exclude_identifier=False, names=None, only_identifier=False
):
    return SourceColumn.get_for_source(
        source=source, exclude_identifier=exclude_identifier,
        names=names, only_identifier=only_identifier
    )


@register.simple_tag(takes_context=True)
def navigation_resolve_menu(context, name, source=None, sort_results=None):
    return _navigation_resolve_menu(
        context=context, name=name, source=source, sort_results=sort_results
    )


@register.simple_tag(takes_context=True)
def navigation_resolve_menus(context, names, source=None, sort_results=None):
    result = []

    for name in names.split(','):
        result.extend(
            _navigation_resolve_menu(
                context=context, name=name, source=source, sort_results=sort_results
            )
        )

    return result


@register.simple_tag(takes_context=True)
def navigation_source_column_get_sort_icon(context, column):
    if column:
        result = column.get_sort_icon(context=context)
        return result
    else:
        return ''


@register.simple_tag(takes_context=True)
def navigation_source_column_resolve(context, column):
    if column:
        result = column.resolve(context=context)
        return result
    else:
        return ''



from django.contrib.auth.models import Group
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.generics import (
    AddRemoveView, MultipleObjectDeleteView, SingleObjectCreateView,
    SingleObjectDetailView, SingleObjectEditView, SingleObjectListView
)

from ..icons import icon_group_setup
from ..links import link_group_create
from ..permissions import (
    permission_group_create, permission_group_delete, permission_group_edit,
    permission_group_view, permission_user_edit
)
from ..querysets import get_user_queryset


class GroupCreateView(SingleObjectCreateView):
    extra_context = {'title': _('Create new group')}
    fields = ('name',)
    model = Group
    post_action_redirect = reverse_lazy(
        viewname='user_management:group_list'
    )
    view_permission = permission_group_create

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class GroupDeleteView(MultipleObjectDeleteView):
    error_message = _('Error deleting group "%(instance)s"; %(exception)s')
    model = Group
    object_permission = permission_group_delete
    pk_url_kwarg = 'group_id'
    post_action_redirect = reverse_lazy(
        viewname='user_management:group_list'
    )
    success_message_single = _('Group "%(object)s" deleted successfully.')
    success_message_singular = _('%(count)d group deleted successfully.')
    success_message_plural = _('%(count)d groups deleted successfully.')
    title_single = _('Delete group: %(object)s.')
    title_singular = _('Delete the %(count)d selected group.')
    title_plural = _('Delete the %(count)d selected groups.')


class GroupDetailView(SingleObjectDetailView):
    fields = ('name',)
    model = Group
    object_permission = permission_group_view
    pk_url_kwarg = 'group_id'

    def get_extra_context(self, **kwargs):
        return {
            'object': self.object,
            'title': _('Details of group: %s') % self.object
        }


class GroupEditView(SingleObjectEditView):
    fields = ('name',)
    model = Group
    object_permission = permission_group_edit
    pk_url_kwarg = 'group_id'
    post_action_redirect = reverse_lazy(
        viewname='user_management:group_list'
    )

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Edit group: %s') % self.object,
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class GroupListView(SingleObjectListView):
    model = Group
    object_permission = permission_group_view

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_group_setup,
            'no_results_main_link': link_group_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'User groups are organizational units. They should '
                'mirror the organizational units of your organization. '
                'Groups can\'t be used for access control. Use roles '
                'for permissions and access control, add groups to '
                'them.'
            ),
            'no_results_title': _('There are no user groups'),
            'title': _('Groups'),
        }


class GroupUsersView(AddRemoveView):
    main_object_method_add_name = 'users_add'
    main_object_method_remove_name = 'users_remove'
    main_object_model = Group
    main_object_permission = permission_group_edit
    main_object_pk_url_kwarg = 'group_id'
    secondary_object_permission = permission_user_edit
    list_available_title = _('Available users')
    list_added_title = _('Group users')

    def get_actions_extra_kwargs(self):
        return {'_event_actor': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _('Users of group: %s') % self.main_object,
        }

    def get_list_added_queryset(self):
        return self.main_object.get_users(
            permission=permission_user_edit, user=self.request.user
        )

    def get_secondary_object_source_queryset(self):
        return get_user_queryset(user=self.request.user)



import json

from django import forms
from django.db.models import Model
from django.db.models.query import QuerySet
from django.forms.formsets import formset_factory
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _

from mayan.apps.templating.fields import ModelTemplateField
from mayan.apps.views.forms import DynamicModelForm, FilteredSelectionForm

from .classes import WorkflowAction
from .fields import WorfklowImageField
from .models import (
    Workflow, WorkflowInstance, WorkflowState, WorkflowStateAction,
    WorkflowTransition
)


class WorkflowActionSelectionForm(forms.Form):
    klass = forms.ChoiceField(
        choices=(), help_text=_('The action type for this action entry.'),
        label=_('Action'), widget=forms.Select(attrs={'class': 'select2'})
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['klass'].choices = WorkflowAction.get_choices()


class WorkflowForm(forms.ModelForm):
    class Meta:
        fields = ('label', 'internal_name', 'auto_launch')
        model = Workflow


class WorkflowMultipleSelectionForm(FilteredSelectionForm):
    class Meta:
        allow_multiple = True
        field_name = 'workflows'
        label = _('Workflows')
        required = False
        widget_attributes = {'class': 'select2'}


class WorkflowStateActionDynamicForm(DynamicModelForm):
    class Meta:
        fields = ('label', 'when', 'enabled', 'condition', 'action_data')
        model = WorkflowStateAction
        widgets = {'action_data': forms.widgets.HiddenInput}

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request')
        self.action_path = kwargs.pop('action_path')
        super().__init__(*args, **kwargs)
        if self.instance.action_data:
            for key, value in json.loads(s=self.instance.action_data).items():
                self.fields[key].initial = value

        self.fields['condition'] = ModelTemplateField(
            initial_help_text=self.fields['condition'].help_text,
            label=self.fields['condition'].label, model=WorkflowInstance,
            model_variable='workflow_instance', required=False
        )

    def clean(self):
        data = super().clean()

        # Consolidate the dynamic fields into a single JSON field called
        # 'action_data'.
        action_data = {}

        for field_name, field_data in self.schema['fields'].items():
            action_data[field_name] = data.pop(
                field_name, field_data.get('default', None)
            )
            if isinstance(action_data[field_name], QuerySet):
                # Flatten the queryset to a list of ids
                action_data[field_name] = list(
                    action_data[field_name].values_list('id', flat=True)
                )
            elif isinstance(action_data[field_name], Model):
                # Store only the ID of a model instance
                action_data[field_name] = action_data[field_name].pk

        data['action_data'] = action_data
        data = import_string(dotted_path=self.action_path).clean(
            form_data=data, request=self.request
        )
        data['action_data'] = json.dumps(obj=action_data)

        return data


class WorkflowStateForm(forms.ModelForm):
    class Meta:
        fields = ('initial', 'label', 'completion')
        model = WorkflowState


class WorkflowTransitionForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        workflow = kwargs.pop('workflow')
        super().__init__(*args, **kwargs)
        self.fields[
            'origin_state'
        ].queryset = self.fields[
            'origin_state'
        ].queryset.filter(workflow=workflow)

        self.fields[
            'destination_state'
        ].queryset = self.fields[
            'destination_state'
        ].queryset.filter(workflow=workflow)

        self.fields['condition'] = ModelTemplateField(
            initial_help_text=self.fields['condition'].help_text,
            label=self.fields['condition'].label, model=WorkflowInstance,
            model_variable='workflow_instance', required=False
        )

    class Meta:
        fields = ('label', 'origin_state', 'destination_state', 'condition')
        model = WorkflowTransition


class WorkflowTransitionTriggerEventRelationshipForm(forms.Form):
    namespace = forms.CharField(
        label=_('Namespace'), required=False,
        widget=forms.TextInput(attrs={'readonly': 'readonly'})
    )
    label = forms.CharField(
        label=_('Label'), required=False,
        widget=forms.TextInput(attrs={'readonly': 'readonly'})
    )
    relationship = forms.ChoiceField(
        label=_('Enabled'),
        widget=forms.RadioSelect(), choices=(
            ('no', _('No')),
            ('yes', _('Yes')),
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['namespace'].initial = self.initial['event_type'].namespace
        self.fields['label'].initial = self.initial['event_type'].label

        relationship = self.initial['transition'].trigger_events.filter(
            event_type=self.initial['event_type'],
        )

        if relationship.exists():
            self.fields['relationship'].initial = 'yes'
        else:
            self.fields['relationship'].initial = 'no'

    def save(self):
        relationship = self.initial['transition'].trigger_events.filter(
            event_type=self.initial['event_type'],
        )

        if self.cleaned_data['relationship'] == 'no':
            relationship.delete()
        elif self.cleaned_data['relationship'] == 'yes':
            if not relationship.exists():
                self.initial['transition'].trigger_events.create(
                    event_type=self.initial['event_type'],
                )


WorkflowTransitionTriggerEventRelationshipFormSet = formset_factory(
    form=WorkflowTransitionTriggerEventRelationshipForm, extra=0
)


class WorkflowInstanceTransitionSelectForm(forms.Form):
    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user')
        workflow_instance = kwargs.pop('workflow_instance')
        super().__init__(*args, **kwargs)
        self.fields[
            'transition'
        ].queryset = workflow_instance.get_transition_choices(_user=user)

    transition = forms.ModelChoiceField(
        help_text=_('Select a transition to execute in the next step.'),
        label=_('Transition'), queryset=WorkflowTransition.objects.none()
    )


class WorkflowPreviewForm(forms.Form):
    workflow = WorfklowImageField()

    def __init__(self, *args, **kwargs):
        instance = kwargs.pop('instance', None)
        super().__init__(*args, **kwargs)
        self.fields['workflow'].initial = instance



from mayan.apps.appearance.classes import Icon
from mayan.apps.permissions.icons import icon_permission

icon_acl_delete = Icon(driver_name='fontawesome', symbol='times')
icon_acl_list = Icon(driver_name='fontawesome', symbol='lock')
icon_acl_new = Icon(
    driver_name='fontawesome-dual', primary_symbol='lock',
    secondary_symbol='plus'
)
icon_acl_permissions = icon_permission
icon_global_acl_list = Icon(
    driver_name='fontawesome-dual', primary_symbol='lock',
    secondary_symbol='globe'
)



import logging

from django.contrib.auth import get_user_model
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import (
    menu_multi_item, menu_object, menu_tools, menu_user
)
from mayan.apps.events.classes import ModelEventType
from mayan.apps.navigation.classes import Separator

from .classes import AuthenticationBackend
from .events import (
    event_user_impersonation_ended, event_user_impersonation_started,
    event_user_logged_in, event_user_logged_out
)
from .handlers import handler_user_logged_in, handler_user_logged_out
from .links import (
    link_logout, link_password_change, link_user_impersonate_form_start,
    link_user_impersonate_start, link_user_multiple_set_password,
    link_user_set_password
)
from .permissions import permission_users_impersonate

logger = logging.getLogger(name=__name__)


class AuthenticationApp(MayanAppConfig):
    app_namespace = 'authentication'
    app_url = 'authentication'
    has_tests = True
    name = 'mayan.apps.authentication'
    verbose_name = _('Authentication')

    def ready(self):
        super().ready()

        User = get_user_model()

        AuthenticationBackend.load_modules()
        AuthenticationBackend.cls_initialize()

        ModelEventType.register(
            model=User, event_types=(
                event_user_impersonation_ended,
                event_user_impersonation_started, event_user_logged_in,
                event_user_logged_out
            )
        )

        ModelPermission.register(
            model=User, permissions=(
                permission_users_impersonate,
            )
        )

        menu_multi_item.bind_links(
            links=(link_user_multiple_set_password,),
            sources=('user_management:user_list',)
        )

        menu_object.bind_links(
            links=(
                link_user_impersonate_start, link_user_set_password,
            ), sources=(User,)
        )

        menu_tools.bind_links(
            links=(link_user_impersonate_form_start,)
        )

        menu_user.bind_links(
            links=(
                Separator(), link_password_change, link_logout
            ), position=99
        )

        user_logged_in.connect(
            dispatch_uid='authentication_handler_user_logged_in',
            receiver=handler_user_logged_in,
            sender=User
        )
        user_logged_out.connect(
            dispatch_uid='authentication_handler_user_logged_out',
            receiver=handler_user_logged_out,
            sender=User
        )



from django.db import migrations, models
import django.db.models.deletion
import mayan.apps.storage.classes
import mayan.apps.storage.models


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('permissions', '0004_auto_20191213_0044'),
        ('storage', '0003_auto_20200601_0649'),
    ]

    operations = [
        migrations.CreateModel(
            name='DownloadFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', models.FileField(storage=mayan.apps.storage.classes.DefinedStorageLazy(name='storage__downloadfile'), upload_to=mayan.apps.storage.models.download_file_upload_to, verbose_name='File')),
                ('filename', models.CharField(db_index=True, max_length=255, verbose_name='Filename')),
                ('datetime', models.DateTimeField(auto_now_add=True, verbose_name='Date time')),
                ('object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('label', models.CharField(db_index=True, max_length=192, verbose_name='Label')),
                ('content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('permission', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='permissions.StoredPermission', verbose_name='Permission')),
            ],
            options={
                'verbose_name': 'Download file',
                'verbose_name_plural': 'Download files',
            },
        ),
    ]



from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.mixins.document_file_mixins import DocumentFileLinkTestMixin

from ..links import link_document_file_parsing_errors_list
from ..permissions import permission_document_file_parse


class DocumentFileContentLinkTestCase(
    DocumentFileLinkTestMixin, GenericDocumentViewTestCase
):
    def test_document_file_parsing_errors_list_link_no_permission(self):
        resolved_link = self._resolve_test_document_file_link(
            test_link=link_document_file_parsing_errors_list
        )
        self.assertEqual(resolved_link, None)

    def test_document_file_parsing_errors_list_link_with_access(self):
        self.grant_access(
            obj=self.test_document_file,
            permission=permission_document_file_parse
        )
        resolved_link = self._resolve_test_document_file_link(
            test_link=link_document_file_parsing_errors_list
        )
        self.assertNotEqual(resolved_link, None)
